from typing import Optional
from longchain.core.agentaction import AgentAction
from longchain.core.dataclasses import Message, NextAction, PathResult


class MessageAgentAction(AgentAction):
    def __init__(self, message: str, name: Optional[str] = None, icon_url: Optional[str] = None):
        self.message = message
        self.name = name
        self.icon_url = icon_url

    async def run(self, path, player, player_actions):
        return PathResult(next_action=self.next_action, new_path_id=None, messages=[Message(text=self.message, interaction_id=player.interaction_id, icon=self.icon_url, name=self.name)])


from collections.abc import Sequence
from longchain.core.actionresolver import ActionResolver
from longchain.core.dataclasses import Message, PathResult, Player, PlayerAction
from longchain.core.agentaction import AgentAction

class Path:
    def __init__(self, id: str, display_name: str, icon: str, starts_without_player_action: bool, action_resolver: ActionResolver):
        self.id = id
        self.display_name = display_name
        self.icon = icon
        self.starts_without_player_action = starts_without_player_action
        self.action_resolver = action_resolver

    async def tick(self, player: Player, player_actions: Sequence[PlayerAction]) -> PathResult:
        try:
            action: AgentAction = await self.action_resolver.tick(player, player_actions)
        except Exception as e:
            return PathResult(is_error=True, next_action='player', new_path_id=None, messages=[Message(text=f"An error occurred while attempting to get an action with {self.action_resolver.__class__.__name__}: {e}", interaction_id=player.interaction_id)])
        try:
            path_result = await action.run(self, player, player_actions)
        except Exception as e:
            return PathResult(is_error=True, next_action='player', new_path_id=None, messages=[Message(text=f"An error occurred while attempting to run a(n) {action.__class__.__name__}: {e}", interaction_id=player.interaction_id)])
        return path_result


from collections.abc import Sequence
import json
import os
from longchain.core.dataclasses import Player
from longchain.core.datastore import Datastore


class JsonFileDatastore(Datastore):
    def __init__(self, path: str):
        self.path = path
        self.players: list[Player] = []
        # try to load the file, and if it doesn't exist, create it
        file_data = ""
        try:
            with open(self.path, 'r') as f:
                file_data = f.read()
        except FileNotFoundError:
            print(f"File not found: {os.path.abspath(self.path)}. Creating...")
            # make sure the directory exists
            os.makedirs(os.path.dirname(self.path), exist_ok=True)
            with open(self.path, 'w') as f:
                f.write('[]')
                file_data = "[]"
        try:
            self.players = [Player(**player_data) for player_data in json.loads(file_data)]
        except json.JSONDecodeError:
            print(f"Error reading JSON from {os.path.abspath(self.path)}")
            raise            
        
    def save_player(self, player):
        # remove the player if it already exists
        self.remove_player(player.id)
        self.players.append(player)
        with open(self.path, 'w') as f:
            f.write(json.dumps([player.__dict__ for player in self.players]))

    def get_player_by_id(self, player_id):
        for player in self.players:
            if player.id == player_id:
                return player
        return None

    def get_player_by_interaction_id(self, interaction_id):
        for player in self.players:
            if player.interaction_id == interaction_id:
                return player
        return None

    def remove_player(self, player_id):
        self.players = [player for player in self.players if player.id != player_id]
        with open(self.path, 'w') as f:
            f.write(json.dumps([player.__dict__ for player in self.players]))


from abc import ABC, abstractmethod
from collections.abc import Sequence, Callable, Awaitable
from typing import Never

from longchain.core.dataclasses import Message, Player, PlayerAction

class Messager(ABC):
    @abstractmethod
    async def send_messages(self, messages: Sequence[Message]):
        pass
    @abstractmethod
    async def start(self, tick: Callable[[Player, Sequence[PlayerAction]], Awaitable[None]], start_interaction: Callable[[Player], Awaitable[None]]) -> Never:
        # this function is called and should never return. it needs to configure itself to call tick whenever
        # a player takes an action and start_interaction whenever a new player starts an interaction.
        # it also needs to handle verifying that the player isn't trying to start a second simultaneous interaction
        pass



from longchain.core.dataclasses import Message
from longchain.core.path import Path
from longchain.core.quest import Quest
from longchain.impl.actionresolver.sequential import SequentialActionResolver
from longchain.impl.agentaction.end import ChangePathAction
from longchain.impl.agentaction.message import MessageAgentAction
from longchain.impl.datastore.jsonfile import JsonFileDatastore
from longchain.impl.messager.slack import SlackMessager
import os

if "ENVIRONMENT" not in os.environ or os.environ["ENVIRONMENT"] != "production":
    import dotenv
    dotenv.load_dotenv(override=True)

ENV_VARS_REQUIRED = ["ENVIRONMENT", "SLACK_BOT_TOKEN", "SLACK_APP_TOKEN", "HOME_CHANNEL_ID", "DATA_FILEPATH", "BOT_USER_ID", "OPENAI_API_KEY", "OPENAI_API_URL", "ST_ADMINS"]
if not all([os.environ.get(var) for var in ENV_VARS_REQUIRED]):
    raise Exception(f"Missing the following environment variables: {', '.join([var for var in ENV_VARS_REQUIRED if not var in os.environ])}")

datastore = JsonFileDatastore(os.environ["DATA_FILEPATH"])

quest = Quest(
    name="Sock Thief",
    paths=[
        Path(
            id="welcome",
            starts_without_player_action=True,
            display_name="Welcome",
            icon="https://example.com/icon.png",
            action_resolver=SequentialActionResolver(
                MessageAgentAction("A cloaked figure approaches you. They look like they're trying to be sneaky, but they're not doing a very good job of it."),
                ChangePathAction("stocking_initial")
            )
        ),
        Path(
            starts_without_player_action=True,
            id="completed",
            display_name="Completed",
            icon="https://example.com/icon.png",
            action_resolver=SequentialActionResolver(
                MessageAgentAction("You have completed the quest!"),
            )
        )
    ],
    message_sender=SlackMessager(
        bot_token=os.environ["SLACK_BOT_TOKEN"],
        app_token=os.environ["SLACK_APP_TOKEN"],
        start_path="welcome",
        datastore=datastore,
        active_channel=os.environ["HOME_CHANNEL_ID"],
        reset_user_command="reset",
        admins=os.environ["ST_ADMINS"].split(',')
        ),
    datastore=datastore
)


from abc import ABC, abstractmethod
from typing import Any, Optional

from longchain.core.dataclasses import Player

class Datastore(ABC):
    @abstractmethod
    def get_player_by_id(self, player_id: Any) -> Optional[Player]:
        pass
    
    @abstractmethod
    def get_player_by_interaction_id(self, interaction_id: Any) -> Optional[Player]:
        pass

    @abstractmethod
    def remove_player(self, player_id: Any):
        pass

    @abstractmethod
    def save_player(self, player: Player):
        pass


from longchain.core.agentaction import AgentAction
from longchain.core.dataclasses import PathResult


class ChangePathAction(AgentAction):
    async def __init__(self, dest: str) -> None:
        self.next_action = "player"
        self.dest = dest
        
    async def run(self, path, player, player_actions):
        return PathResult(next_action=self.next_action, new_path_id=self.dest, messages=[])


from abc import abstractmethod, ABC
from collections.abc import Sequence

from longchain.core.dataclasses import NextAction, PathResult, Player, PlayerAction
from longchain.core.path import Path

class AgentAction(ABC):
    _next_action: NextAction = 'player'
    # this gurantees to ActionResolvers that they can set the next action if they want to
    @property
    def next_action(self) -> NextAction:
        return self._next_action
    @next_action.setter
    def next_action(self, value: NextAction):
        self._next_action = value

    @abstractmethod
    async def run(self, path: Path, player: Player, player_actions: Sequence[PlayerAction]) -> PathResult:
        pass


from collections.abc import Sequence
from longchain.core.dataclasses import Message, PathResult, Player, PlayerAction
from longchain.core.datastore import Datastore
from longchain.core.messager import Messager
from longchain.core.path import Path

class Quest:
    def __init__(self, name: str, paths: Sequence[Path], message_sender: Messager, datastore: Datastore):
        self.name = name
        self.first_path = paths[0].id
        self.paths = {path.id: path for path in paths}
        self.messager = message_sender
        self.datastore = datastore
        if not paths[0].starts_without_player_action:
            raise ValueError("The first path in a quest must start without a player action")
    
    async def tick(self, player: Player, player_actions: Sequence[PlayerAction]):
        ready_to_end = False
        messages: list[Message] = []
        while not ready_to_end: # loop until a path asks for a player action
            current_path = self.paths[player.current_path] # get the path the player is on
            path_result: PathResult = await current_path.tick(player, player_actions) # run that path
            messages.extend(path_result.messages) # add any messages it generated to the queue
            ready_to_end = path_result.next_action == 'player' # if the path asks for a player action, we're done
            if path_result.new_path_id: # if the path changed:
                # check that the new path exists
                if path_result.new_path_id not in self.paths:
                    messages.append(Message(text=f"Error: tried to switch to new path {path_result.new_path_id}, but it does not exist", interaction_id=player.interaction_id))
                    ready_to_end = True
                    break
                player.current_path = path_result.new_path_id # update the player's current path
                if self.paths[player.current_path].starts_without_player_action:
                    # this path needs a player action to start, so we don't care if the previous path asked for anther agent turn
                    ready_to_end = True
        await self.messager.send_messages(messages)
        self.datastore.save_player(player)
    
    async def start_interaction(self, player: Player):
        player.current_path = self.first_path
        await self.tick(player, [])

    async def run(self):
        await self.messager.start(self.tick, self.start_interaction)


from typing import Literal
from longchain.core.agentaction import AgentAction
from longchain.core.actionresolver import ActionResolver
from longchain.core.dataclasses import Player

# Calls all its arguments in order, then goes to the player (by default.)
# Usually, should end with a ChangePathAction.
class SequentialActionResolver(ActionResolver):
    def __init__(self, *actions: AgentAction, on_end: Literal['go_to_player', 'repeat_from_start'] = 'go_to_player'):
        self.actions = actions
        self.on_end = on_end
    async def tick(self, player, player_actions) -> AgentAction:
        action_number = player.path_states[player.current_path]["action_number"]
        if action_number >= len(self.actions):
            raise ValueError("No more actions to run")
        action = self.actions[action_number]
        player.path_states[player.current_path]["action_number"] += 1
        action.next_action = "agent" if action_number < len(self.actions) - 1 or self.on_end == 'repeat_from_start' else "player"
        # call the next agent (ususally us) unless this is the last action and we're supposed to go to the player
        return action


quest = Quest(
    name="test",
    paths=[
        Path(
            id="smallberry_initial",
            display_name="Baker Smallberry",
            icon="example.com/smallberry.png",
            action_reolver=LlmActionResolver(
                id="smallberry_initial_llm",
                system_prompt="example prompt",
                use_history_from=["smallberry_initial_llm"],
                clear_history_on_end=False,
                endable_paths=[EndablePath(path_id="traveling_there", scenario="When the player has agreed to the quest")],
            )
        ),
        Path(
            id="travelling_there",
            display_name="travelling...",
            icon="example.com/travelling.png",
            action_resolver=SequentialActionResolver(
                MessageAction(
                    message="You travel over the hill and through the woods" # Resolvable<string>
                ),
                EndAction(
                    dest=lambda p: random.choices(["travelling_there_2", "bandit_encounter"], [0.75, 0.25]) # Resolvable<PathId>
                )
            )
        ),
        Path(
            id="bandit_encounter",
            display_name="Bandit Encounter",
            icon="example.com/bandit.png",
            action_resolver=LlmActionResolver(
                id="bandit_encounter_llm",
                system_prompt="example prompt",
                use_history_from=["bandit_encounter_llm"],
                clear_history_on_end=True,
                agent_actions=[
                    LlmTool(
                        name="take_gold",
                        description="Take some amount of gold from the player",
                        params=[
                            LlmToolParam(
                                name="num_gold",
                                description="The number of gold to take",
                                type="int"
                                min=0,
                                max=100
                            )
                        ],
                        available=lambda ctx: bag.get_stake(player).gold > 0 and ctx.times_this_called < 1,
                        func=lambda ctx: LlmToolResult(action=bag.TakeItemAction(item="gp", num=ctx.params["num_gold"]), description=f"The bandits take {ctx.params['num_gold']} gold from you", reprompt=True)
                    ),
                    LlmTool(
                        name="end_bandit_encounter",
                        description="End the bandit encounter. Call when the bandits leave, either because the player defeated them, because they're done stealing, or because the player scares them off",
                        params=[],
                        available= ctx.num_actions_taken > 2,
                        action=lambda ctx: LlmToolResult(action=EndAction(dest="travelling_there_2"), description="The bandits leave", reprompt=False)
                    )
                ]
            )
        )
    ]
)

class Path:
    def tick(player: Player) -> PathResult:
        action: Action = self.action_resolver.tick(player)
        path_result = action.run(self, player)
        return path_result

class Quest:
    def tick(player):
        path_result: PathResult = self.get_path(player.current_path).tick(player)
        slack.send_messages(path_result.messages)
        if path_result.new_path_id:
            player.current_path = path_result.new_path_id
    
await quest.run()

# so:
# a new message calls quest.tick with that player
# quest.tick calls the current path's tick method
# the current path's tick method calls the action resolver's tick method
# the action resolver's tick method returns an action
# the action's run method is called with the path and player
# the action's run method might have side effects, like changing the player's state, or changing their inventory
# the action's run method returns a path result
# the path result is returned to the quest's tick method
# the quest's tick method deals with the path result by changing the path if needed and sending the messages to slack
# the quest's tick method returns


from abc import ABC, abstractmethod
from collections.abc import Sequence

from longchain.core.agentaction import AgentAction
from longchain.core.dataclasses import Player, PlayerAction


class ActionResolver(ABC):
    @abstractmethod
    async def tick(self, player: Player, player_actions: Sequence[PlayerAction]) -> AgentAction:
        pass


from collections.abc import Awaitable, Callable, Sequence
import re
from typing import Never
from longchain.core.dataclasses import Player, PlayerAction
from longchain.core.datastore import Datastore
from longchain.core.messager import Messager

try:
    from slack_bolt.app.async_app import AsyncApp
    from slack_bolt.adapter.socket_mode.async_handler import AsyncSocketModeHandler
    from slack_sdk.web.async_client import AsyncWebClient
except ImportError:
    raise ImportError("The `slack` messager is not available when the `slack` extra is not installed.")

class SlackMessager(Messager):
    def __init__(self, bot_token: str, app_token: str, start_path: str, datastore: Datastore, active_channel: str, reset_user_command: str, admins: list[str], workspace_name: str = 'hackclub'):
        self.start_path = start_path
        self.datastore = datastore
        self.active_channel = active_channel
        self.reset_user_command = reset_user_command
        self.admins = admins
        self.app = AsyncApp(token=bot_token)
        self.handler = AsyncSocketModeHandler(app=self.app, app_token=app_token)
        self.workspace_name = workspace_name

    async def send_messages(self, messages):
        for i, message in enumerate(messages):
            player = self.datastore.get_player_by_interaction_id(message.interaction_id)
            if player is None:
                print(f"ERROR: Player not found for message {message.interaction_id}")
                # this should be unreachable, but just in case
                continue
            blocks = [
                {
                    "type": "section",
                    "text": {
                        "type": "mrkdwn",
                        "text": message.text
                    }
                }
            ]
            if i == len(messages) - 1:
                blocks.append({
                    "type": "context",
                    "elements": [
                        {
                            "type": "mrkdwn",
                            "text": "It's your turn to send a message now!"
                        }
                    ]
                })
            await self.app.client.chat_postMessage(channel=player.messager_state["channel"], text=message.text, thread_ts=message.interaction_id, username=message.name, icon_url=message.icon, blocks=blocks)

    async def handle_thread_message(self, client, body, tick: Callable[[Player, Sequence[PlayerAction]], Awaitable[None]]):
        player = self.datastore.get_player_by_interaction_id(body["event"]["thread_ts"])
        if player is None:
            # this thread is not a game thread, ignore it
            return
        if player.id != body["event"]["user"]:
            await client.chat_postEphemeral(channel=body["event"]["channel"], thread_ts=body["event"]["thread_ts"], user=body["event"]["user"], text=f"Hey, you're not <@{player.id}>! You're being ignored, you hear me? Ignored!")
            return
        if player.messager_state["last_unprocessed_message"] < float(body["event"]["ts"]) or player.messager_state["last_processed_message"] is not None and player.messager_state["last_processed_message"] < float(body["event"]["ts"]):
            # okay that logic is a bit of a doozy but:
            # if the last unprocessed message is older than this one, then we're still processing the last one
            # or, if the last processed message is older than this one, then we're still processing another one
            # in either case, we should ignore this message
            await client.reactions_add(channel=body["event"]["channel"], name="tw_x", timestamp=body["event"]["ts"])
            await client.chat_postEphemeral(channel=body["event"]["channel"], thread_ts=body["event"]["thread_ts"], user=body["event"]["user"], text="I'm still processing your last message. This message was ignored. Please don't send multiple messages in a row.")
            if player.messager_state["last_processed_message"] is None or player.messager_state["last_processed_message"] < float(body["event"]["ts"]):
                # this counts as processing the message, so update the last processed message time if it's newer
                player.messager_state["last_processed_message"] = float(body["event"]["ts"])
            return
        player.messager_state["last_unprocessed_message"] = float(body["event"]["ts"])
        # okay by this point we should be good to actually process the message
        # with the gurantees that this player is real, the plaayer is the one who sent
        # the message, and that we're not processing another of the same player's message already
        # start by creating player actions
        # for now, we'll just assume the player is sending a message
        # TODO: support buttons, somehow
        player_actions = [PlayerAction(name="say", data={"message": body["event"]["text"]})]
        # finally, tick the quest
        await tick(player, player_actions)
        
    async def start(self, tick, start_interaction) -> Never:
        @self.app.event("app_mention")
        async def handle_app_mention(client, body, say, ack):
            await ack()
            if "thread_ts" in body["event"]:
                pass # we were pinged in a thread, the other handler'll catch it
            elif body["event"]["channel"] == self.active_channel:
                player = self.datastore.get_player_by_id(body["event"]["user"])
                if player is not None:
                    await client.chat_postEphemeral(channel=body["event"]["channel"], user=body["event"]["user"], text=f"You're already in a conversation with me. Please continue that conversation <https://{self.workspace_name}.slack.com/archives/{self.active_channel}/p{str(player.interaction_id).replace('.','')}|here>.")
                    return
                player_name = await client.users_info(user=body["event"]["user"])
                player = Player(id=body["event"]["user"], current_path=self.start_path, name=player_name["user"]["profile"]["display_name"], interaction_id=body["event"]["ts"], path_states={}, messager_state={"channel": body["event"]["channel"], "last_unprocessed_message": None, "last_processed_message": float(body["event"]["ts"])})
                await start_interaction(player)
            else:
                await say(f"I can't speak here. Try pinging me in <#{self.active_channel}>.")
        
        username_re = re.compile(r"<@(.*)\|.*>")

        @self.app.command(self.reset_user_command)
        async def handle_reset_user_command(ack, command, client, respond):
            await ack()
            print(f"Got reset user command from {command['user_id']}")
            if command["user_id"] not in self.admins:
                await respond(f"Sorry, you're not allowed to do that.", responce_type="ephemeral")
                return
            
            mo = username_re.match(command["text"])
            if mo is None or len(mo.groups()) != 1:
                await respond(f"That didn't parse properly. Make sure you're @ing a user.", response_type="ephemeral")
                return
            
            user_id_to_reset = mo.group(1)
            
            player = self.datastore.get_player_by_id(user_id_to_reset)
            if player is None:
                await respond(f"Player not found. Why are you running this command?", response_type="ephemeral")
                return
            
            self.datastore.remove_player(user_id_to_reset)
            print(f"Player {user_id_to_reset} reset.")
            await respond(f"Player <@{player.id}> reset.", response_type="ephemeral")
        
        @self.app.message()
        async def handle_message(client, body, say):
            if "thread_ts" in body["event"]:
                await self.handle_thread_message(client, body, tick)

        await self.handler.start_async()
        while True: # unreachable, the above call never returns
            pass


from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from typing import Any, Literal, Optional


@dataclass
class Message:
    text: str
    interaction_id: str
    icon: Optional[str] = None
    name: Optional[str] = None

NextAction = Literal['player', 'path']
@dataclass
class PathResult:
    next_action: NextAction
    new_path_id: Optional[str]
    messages: Sequence[Message]
    is_error: bool = False

@dataclass
class Player:
    id: str
    current_path: str
    name: str
    interaction_id: str
    path_states: Mapping[str, dict]
    messager_state: dict

@dataclass
class PlayerAction:
    name: str
    data: dict